@@ -1,6 +1,13 @@
 # coding=utf-8
 from __future__ import division
+import os
+import sys
+import json
 import collections
+try:  # check if we are in IronPython
+    import cPickle as pickle
+except ImportError:  # wea are in cPython
+    import pickle
 
 from ladybug.legend import Legend, LegendParameters, LegendParametersCategorized
 from ladybug.graphic import GraphicContainer
@@ -127,6 +134,16 @@ class _VisualizationBase(object):
                 'object user_data. Got {}.'.format(type(value))
         self._user_data = value
 
+    def duplicate(self):
+        """Get a copy of this object."""
+        return self.__copy__()
+
+    def __copy__(self):
+        new_obj = _VisualizationBase(self.identifier)
+        new_obj._display_name = self._display_name
+        new_obj._user_data = None if self.user_data is None else self.user_data.copy()
+        return new_obj
+
 
 class VisualizationSet(_VisualizationBase):
     """A visualization set containing analysis and context geometry to be visualized.
@@ -187,6 +204,54 @@ class VisualizationSet(_VisualizationBase):
             new_obj.user_data = data['user_data']
         return new_obj
 
+    @classmethod
+    def from_file(cls, hb_file):
+        """Initialize a VisualizationSet from a JSON or pkl file, auto-sensing the type.
+
+        Args:
+            hb_file: Path to either a VisualizationSet JSON or pkl file.
+        """
+        # sense the file type from the first character to avoid maxing memory with JSON
+        # this is needed since queenbee overwrites all file extensions
+        with open(hb_file) as inf:
+            try:
+                first_char = inf.read(1)
+                is_json = True if first_char == '{' else False
+            except UnicodeDecodeError:  # definitely a pkl file
+                is_json = False
+        # load the file using either JSON pathway or pkl
+        if is_json:
+            return cls.from_json(hb_file)
+        return cls.from_pkl(hb_file)
+
+    @classmethod
+    def from_json(cls, json_file):
+        """Initialize a VisualizationSet from a JSON file.
+
+        Args:
+            json_file: Path to VisualizationSet JSON file.
+        """
+        assert os.path.isfile(json_file), 'Failed to find %s' % json_file
+        if (sys.version_info < (3, 0)):
+            with open(json_file) as inf:
+                data = json.load(inf)
+        else:
+            with open(json_file, encoding='utf-8') as inf:
+                data = json.load(inf)
+        return cls.from_dict(data)
+
+    @classmethod
+    def from_pkl(cls, pkl_file):
+        """Initialize a Model from a pkl file.
+
+        Args:
+            pkl_file: Path to pkl file.
+        """
+        assert os.path.isfile(pkl_file), 'Failed to find %s' % pkl_file
+        with open(pkl_file, 'rb') as inf:
+            data = pickle.load(inf)
+        return cls.from_dict(data)
+
     @classmethod
     def from_single_analysis_geo(
             cls, identifier, geometry, values, legend_parameters=None,
@@ -265,6 +330,8 @@ class VisualizationSet(_VisualizationBase):
             geos_list = list(self._geometry)
             geos_list.insert(insert_index, geometry)
             self._geometry = tuple(geos_list)
+        self._min_point = None
+        self._max_point = None
 
     def remove_geometry(self, geo_index):
         """Remove a geometry object from this VisualizationSet.
@@ -278,6 +345,8 @@ class VisualizationSet(_VisualizationBase):
         geos_list = list(self._geometry)
         geos_list.pop(geo_index)
         self._geometry = tuple(geos_list)
+        self._min_point = None
+        self._max_point = None
 
     def check_duplicate_identifiers(self, raise_exception=True, detailed=False):
         """Check that there are no duplicate geometry object identifiers in the set.
@@ -362,6 +431,49 @@ class VisualizationSet(_VisualizationBase):
             base['user_data'] = self.user_data
         return base
 
+    def to_json(self, name, folder, indent=None):
+        """Write VisualizationSet to JSON.
+
+        Args:
+            name: A text string for the name of the JSON file.
+            folder: A text string for the directory where the JSON will be written.
+            indent: A positive integer to set the indentation used in the resulting
+                JSON file. (Default: None).
+        """
+        # create dictionary from the VisualizationSet
+        vs_dict = self.to_dict()
+        # set up a name and folder for the JSON
+        file_name = name if name.lower().endswith('.json') else '{}.json'.format(name)
+        if not os.path.isdir(folder):
+            os.makedirs(folder)
+        vs_file = os.path.join(folder, file_name)
+        # write JSON
+        with open(vs_file, 'w') as fp:
+            json.dump(vs_dict, fp, indent=indent)
+        return vs_file
+
+    def to_pkl(self, name, folder):
+        """Write VisualizationSet to compressed pickle file (pkl).
+
+        Args:
+            name: A text string for the name of the pickle file.
+            folder: A text string for the directory where the pickle file will be
+                written.
+        """
+        # create dictionary from the VisualizationSet
+        vs_dict = self.to_dict()
+        # set up a name and folder for the pkl
+        if name is None:
+            name = self.identifier
+        file_name = name if name.lower().endswith('.pkl') else '{}.pkl'.format(name)
+        if not os.path.isdir(folder):
+            os.makedirs(folder)
+        vs_file = os.path.join(folder, file_name)
+        # write the Model dictionary into a file
+        with open(vs_file, 'wb') as fp:
+            pickle.dump(vs_dict, fp)
+        return vs_file
+
     def _check_geometry(self, geo):
         """Check that the geometry object is valid."""
         assert isinstance(geo, (AnalysisGeometry, ContextGeometry)), 'Expected ' \
@@ -381,6 +493,13 @@ class VisualizationSet(_VisualizationBase):
         """Overwrite .NET ToString."""
         return self.__repr__()
 
+    def __copy__(self):
+        new_geo_objs = tuple(obj.duplicate() for obj in self.geometry)
+        new_obj = VisualizationSet(self.identifier, new_geo_objs)
+        new_obj._display_name = self._display_name
+        new_obj._user_data = None if self.user_data is None else self.user_data.copy()
+        return new_obj
+
     def __len__(self):
         """Return number of geometries on the object."""
         return len(self.geometry)
@@ -549,6 +668,13 @@ class ContextGeometry(_VisualizationBase):
         lb_geos = [d_geo.geometry for d_geo in self.geometry]
         self._min_point, self._max_point = bounding_box(lb_geos)
 
+    def __copy__(self):
+        new_geo_objs = tuple(geo.duplicate() for geo in self.geometry)
+        new_obj = ContextGeometry(self.identifier, new_geo_objs)
+        new_obj._display_name = self._display_name
+        new_obj._user_data = None if self.user_data is None else self.user_data.copy()
+        return new_obj
+
     def __len__(self):
         """Return number of geometries on the object."""
         return len(self.geometry)
@@ -868,6 +994,14 @@ class AnalysisGeometry(_VisualizationBase):
                     'object for AnalysisGeometry. Got {}.'.format(type(geo))
         return (geo_count_0, geo_count_1, geo_count_2)
 
+    def __copy__(self):
+        new_d = tuple(data.duplicate() for data in self.data_sets)
+        new_obj = AnalysisGeometry(
+            self.identifier, self.geometry, new_d, self.active_data, self.display_mode)
+        new_obj._display_name = self._display_name
+        new_obj._user_data = None if self.user_data is None else self.user_data.copy()
+        return new_obj
+
     def __len__(self):
         """Return number of data sets on the object."""
         return len(self.data_sets)
@@ -908,6 +1042,7 @@ class VisualizationMetaData(object):
     Properties:
         * legend_parameters
         * data_type
+        * unit
         * user_data
     """
     __slots__ = ('_legend_parameters', '_data_type', '_unit', '_user_data')
@@ -915,44 +1050,9 @@ class VisualizationMetaData(object):
     def __init__(self, legend_parameters=None, data_type=None, unit=None):
         """Initialize VisualizationMetaData."""
         self._legend_parameters = legend_parameters
-        self._user_data = None
-
-        # set default legend parameters based on input data_type and unit
         self._data_type = data_type
         self._unit = unit
-        if data_type is not None:
-            assert isinstance(data_type, DataTypeBase), \
-                'data_type should be a ladybug DataType. Got {}'.format(type(data_type))
-            if self.legend_parameters.is_title_default:
-                unit = unit if unit else data_type.units[0]
-                data_type.is_unit_acceptable(unit)
-                self.legend_parameters.title = unit if \
-                    self.legend_parameters.vertical \
-                    else '{} ({})'.format(data_type.name, unit)
-            if data_type.unit_descr is not None and \
-                    self.legend_parameters.ordinal_dictionary is None and not \
-                    isinstance(self.legend_parameters, LegendParametersCategorized):
-                self.legend_parameters.ordinal_dictionary = data_type.unit_descr
-                sorted_keys = sorted(data_type.unit_descr.keys())
-                if self.legend.is_min_default:
-                    self.legend_parameters._min = sorted_keys[0]
-                if self.legend.is_max_default:
-                    self.legend_parameters._max = sorted_keys[-1]
-                assert self.legend_parameters._min <= self.legend_parameters._max, \
-                    'Legend min is greater than legend max. {} > {}.'.format(
-                        self.legend_parameters._min, self.legend_parameters._max)
-                if self.legend_parameters.is_segment_count_default:
-                    try:  # try to set the number of segments to align with ordinal text
-                        min_i = sorted_keys.index(self.legend_parameters.min)
-                        max_i = sorted_keys.index(self.legend_parameters.max)
-                        self.legend_parameters.segment_count = \
-                            len(sorted_keys[min_i:max_i + 1])
-                    except IndexError:
-                        pass
-        elif unit and self.legend_parameters.is_title_default:
-            assert isinstance(unit, str), \
-                'Expected string for unit. Got {}.'.format(type(unit))
-            self.legend_parameters.title = unit
+        self._user_data = None
 
     @classmethod
     def from_dict(cls, data):
@@ -1068,16 +1168,54 @@ class VisualizationData(VisualizationMetaData):
     Properties:
         * values
         * legend_parameters
+        * legend
         * data_type
+        * unit
+        * value_colors
         * user_data
     """
-    __slots__ = ('_legend', '_legend_parameters', '_data_type', '_unit', '_user_data')
+    __slots__ = ('_legend',)
 
     def __init__(self, values, legend_parameters=None, data_type=None, unit=None):
         """Initialize VisualizationData."""
         # set up the legend using the values and legend parameters
         VisualizationMetaData.__init__(self, legend_parameters, data_type, unit)
+
+        # assign defaults to the legend parameter using the values and the data type
         self._legend = Legend(values, legend_parameters)
+        if data_type is not None:
+            assert isinstance(data_type, DataTypeBase), \
+                'data_type should be a ladybug DataType. Got {}'.format(type(data_type))
+            if self.legend_parameters.is_title_default:
+                unit = unit if unit else data_type.units[0]
+                data_type.is_unit_acceptable(unit)
+                self.legend_parameters.title = unit if \
+                    self.legend_parameters.vertical \
+                    else '{} ({})'.format(data_type.name, unit)
+            if data_type.unit_descr is not None and \
+                    self.legend_parameters.ordinal_dictionary is None and not \
+                    isinstance(self.legend_parameters, LegendParametersCategorized):
+                self.legend_parameters.ordinal_dictionary = data_type.unit_descr
+                sorted_keys = sorted(data_type.unit_descr.keys())
+                if self.legend.is_min_default:
+                    self.legend_parameters._min = sorted_keys[0]
+                if self.legend.is_max_default:
+                    self.legend_parameters._max = sorted_keys[-1]
+                assert self.legend_parameters._min <= self.legend_parameters._max, \
+                    'Legend min is greater than legend max. {} > {}.'.format(
+                        self.legend_parameters._min, self.legend_parameters._max)
+                if self.legend_parameters.is_segment_count_default:
+                    try:  # try to set the number of segments to align with ordinal text
+                        min_i = sorted_keys.index(self.legend_parameters.min)
+                        max_i = sorted_keys.index(self.legend_parameters.max)
+                        self.legend_parameters.segment_count = \
+                            len(sorted_keys[min_i:max_i + 1])
+                    except IndexError:
+                        pass
+        elif unit and self.legend_parameters.is_title_default:
+            assert isinstance(unit, str), \
+                'Expected string for unit. Got {}.'.format(type(unit))
+            self.legend_parameters.title = unit
 
     @classmethod
     def from_dict(cls, data):
@@ -1122,6 +1260,11 @@ class VisualizationData(VisualizationMetaData):
         """Get the values assigned to the data set."""
         return self._legend.values
 
+    @property
+    def legend_parameters(self):
+        """Get the legend parameters assigned to this data set."""
+        return self._legend._legend_par
+
     @property
     def legend(self):
         """Get the legend assigned to this data set."""
@@ -1161,6 +1304,16 @@ class VisualizationData(VisualizationMetaData):
             base['user_data'] = self.user_data
         return base
 
+    def duplicate(self):
+        """Get a copy of this object."""
+        return self.__copy__()
+
+    def __copy__(self):
+        new_obj = VisualizationData(
+            self.values, self._legend_parameters, self._data_type, self._unit)
+        new_obj._user_data = None if self.user_data is None else self.user_data.copy()
+        return new_obj
+
     def __len__(self):
         """Return length of values on the object."""
         return len(self._legend._values)
